// Hippo Direct Draw
// 2000 (c)
// under construction

#include <stdafx.h>

#include "ddutils.h"
#include "hddraw.h"

//////////////////////////////////////////////////////////////////////////////////
// Inicializace kompletni nulovani
//////////////////////////////////////////////////////////////////////////////////
void HDDraw::InitHDDraw()
{
	HRESULT   rval;

	// initialize member variables
	m_lpDD					= NULL;
	m_lpDDSGDI				= NULL;
	m_lpDDSBack				= NULL;
//	m_dcBack             = NULL;
	m_bFullScreen			= FALSE;
	m_GDIVisible         = TRUE;

	m_Rect.left = 0; m_Rect.top = 0;
	m_Rect.right = 0; m_Rect.bottom = 0;

	// create direct draw object

	rval = DirectDrawCreate(NULL,(LPDIRECTDRAW *) &m_lpDD, NULL);
	if(rval != DD_OK) m_lpDD = NULL;
}


//////////////////////////////////////////////////////////////////////////////////
// standartni konstruktor
//////////////////////////////////////////////////////////////////////////////////
HDDraw::HDDraw()
{
	InitHDDraw();
}

//////////////////////////////////////////////////////////////////////////////////
// ma jteny efekt, jako normalni konstruktor a nasledne volani CreateFullScreen
//////////////////////////////////////////////////////////////////////////////////
HDDraw::HDDraw(HWND hWnd, DWORD Width, DWORD Height, DWORD Color)
{
	InitHDDraw();
	CreateFullScreen(hWnd, Width, Height, Color);
}

//////////////////////////////////////////////////////////////////////////////////
// destruktor - nici nenulove obekty
//////////////////////////////////////////////////////////////////////////////////
HDDraw::~HDDraw(void)
{

//	if(m_dcBack != NULL)
//		ReleaseBackDC();

	if( m_lpDDSBack != NULL ) {
		m_lpDDSBack->Release();
		m_lpDDSBack = NULL;
	}

	if( m_lpDDSGDI != NULL ) {
		m_lpDDSGDI->Release();
		m_lpDDSGDI = NULL;
	}

	if(m_lpDD != NULL) {
		m_lpDD->RestoreDisplayMode();
		m_lpDD->Release();
	}

}

//////////////////////////////////////////////////////////////////////////////////
// prepne do fullscreenu a vytvori predni a zadni buffer - pouziva 16BPP (bytes per pixel)
//////////////////////////////////////////////////////////////////////////////////
BOOL HDDraw::CreateFullScreen(HWND hWnd, DWORD Width, DWORD Height, DWORD color)
{
	if(m_lpDD == NULL) return FALSE;

	HRESULT ddrval;

	DDSURFACEDESC ddsd;
	DDSCAPS ddscaps;

	// first release everything that has been allocated
	if( m_lpDDSBack != NULL ) {
		delete m_lpDDSBack;
		m_lpDDSBack = NULL;
	}

	if( m_lpDDSGDI != NULL ) {
		delete m_lpDDSGDI;
		m_lpDDSGDI = NULL;
	}


	m_bFullScreen = FALSE;

	// Get exclusive mode - spojeni s objektem okna
	ddrval = m_lpDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT);
	if( ddrval != DD_OK ) return FALSE;

	// zmena rozliseni obrazovky
	ddrval = m_lpDD->SetDisplayMode( 640, 480, 16);  // zmena rozliseni - 16-bitova barevna hloubka
   if( ddrval != DD_OK ) 
		return FALSE;

	// Create the primary surface with 1 back buffer - tvorba surfacu
	ddsd.dwSize = sizeof( ddsd );
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
	ddsd.dwBackBufferCount = 1; // pocet pouze pametovych obrazovek - na fillig - 1
	ddrval = m_lpDD->CreateSurface( &ddsd, &m_lpDDSGDI, NULL );
	if( ddrval != DD_OK ) return FALSE;

	FillGDI(color);


	// Get a pointer to the back buffer 
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
   ddrval = m_lpDDSGDI->GetAttachedSurface(&ddscaps, &m_lpDDSBack);
	if( ddrval != DD_OK ) return FALSE;

	Fill(color);

	// pripraveni alpha blittingu
	DDPIXELFORMAT pix;
	DDInitStruct(&pix);
	m_lpDDSGDI->GetPixelFormat(&pix);

	if(InitAlpha(pix) != DD_OK) return FALSE;

	m_bFullScreen = TRUE;
	m_Rect.left = 0; m_Rect.top = 0;
	m_Rect.right = Width; m_Rect.bottom = Height;

	return TRUE;
}

//////////////////////////////////////////////////////////////////////////////////
// prohodi predni a zadni buffer
//////////////////////////////////////////////////////////////////////////////////
HRESULT HDDraw::Flip(BOOL VSync)
{
	HRESULT rval;


	if(m_bFullScreen)
	{
		if(m_FlipLock) {
			rval = m_lpDDSGDI->BltFast(0,0,m_lpDDSBack,&m_Rect,DDBLTFAST_WAIT);
		/*	GetBackDC();
			GetGDIDC();
			BitBlt(m_dcGDI,0,0,640,480,m_dcBack,0,0,SRCCOPY);
			ReleaseGDIDC();
			ReleaseBackDC(); */
		}
		else {
			if( VSync == TRUE )
				rval = m_lpDDSGDI->Flip(NULL, DDFLIP_WAIT);
			else
				rval = m_lpDDSGDI->Flip(NULL, DDFLIP_NOVSYNC);
		}

		if(rval == DDERR_SURFACELOST) Restore();
		if((rval == DD_OK)&&(!m_FlipLock)) m_GDIVisible = !m_GDIVisible;
	}
	return rval;
}

//////////////////////////////////////////////////////////////////////////////////
// vyplni zadni buffer danou barvou
//////////////////////////////////////////////////////////////////////////////////
void HDDraw::Fill(DWORD FillColor)
{
	DDBLTFX ddBltFx;

	ddBltFx.dwSize = sizeof(DDBLTFX);
	ddBltFx.dwFillColor = FillColor;
	m_lpDDSBack->Blt(NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddBltFx);
}

//////////////////////////////////////////////////////////////////////////////////
// vyplni preni buffer danou barvou
//////////////////////////////////////////////////////////////////////////////////
void HDDraw::FillGDI(DWORD FillColor)
{
	DDBLTFX ddBltFx;

	ddBltFx.dwSize = sizeof(DDBLTFX);
	ddBltFx.dwFillColor = FillColor;
	m_lpDDSGDI->Blt(NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddBltFx);
}


//////////////////////////////////////////////////////////////////////////////////
// kdyz se surface ztrati
//////////////////////////////////////////////////////////////////////////////////
void HDDraw::Restore(void)
{
	if(m_bFullScreen) {
		// ve full screenu staci resornou t predn ibuffer, zadni se resotrne automaticky
		if(m_lpDDSGDI != NULL && m_lpDDSGDI->IsLost() != DD_OK)
			m_lpDDSGDI->Restore();
	}
/*	else {
		//  mimo fullscreen se musi oba buffery restornout zvlast
		if(m_lpDDSGDI != NULL && m_lpDDSGDI->IsLost() != DD_OK)
			m_lpDDSGDI->Restore();

		if(m_lpDDSBack != NULL && m_lpDDSBack->IsLost() != DD_OK)
			m_lpDDSBack->Restore();
	}
*/
}

//////////////////////////////////////////////////////////////////////////////////
// !!!!!!! NEFUNGUJE !!!!!!!!
//////////////////////////////////////////////////////////////////////////////////
void HDDraw::FadeToSource(WORD duration)
{
/*	float fadeInc;
	LONG ALPHA;
	DWORD fadeBegin;
	DWORD w,h;
	RECT sRect;
	CDXSurface* tmp;

	// Create a tmp surface the size of the back buffer and
	// copy the back buffer to it.
	w = GetBack()->GetWidth();
	h = GetBack()->GetHeight();
	sRect.left   = 0;
	sRect.top    = 0;
	sRect.right  = w;
	sRect.bottom = h;

	tmp = new CDXSurface(this, w, h, CDXMEM_SYSTEMONLY);
	GetBack()->CDXBLT_Blk(tmp, 0, 0, sRect);

	ALPHA = 0;

	// Start with a completely black screen
	GetBack()->Fill(0);
	Flip(TRUE,FALSE);
		// Continue to brighten the display
	while(ALPHA < 256) {
		fadeBegin = timeGetTime();

		tmp->CDXBLT_Fade(GetBack(),0,0,sRect,(WORD)ALPHA);
		Flip(TRUE,FALSE);

		fadeInc = (float)(duration + 1) / (float)(timeGetTime() - fadeBegin);

		// Do some boundry checking
		if (fadeInc > 256)
			fadeInc = 256;
		else if (fadeInc < 1)
			fadeInc = 1;

		ALPHA += (LONG)(512 / fadeInc);
	}

	// Just to be sure we have a pure copy of the source
	tmp->CDXBLT_Blk(GetBack(), 0, 0, sRect);
	Flip(TRUE,FALSE);
		
	// Cleanup the surface
	delete tmp;
	tmp = NULL;*/
}

//////////////////////////////////////////////////////////////////////////////////
// !!!!!!! NEFUNGUJE !!!!!!!!
//////////////////////////////////////////////////////////////////////////////////
void HDDraw::FadeToBlack(WORD duration)
{
/*	float fadeInc;
	LONG ALPHA;
	DWORD fadeBegin;
	CDXSurface* tmp;
	RECT sRect;
	DWORD w, h;
	POINT pt;

	if (m_BPP == 8)
	{
		// Sorry this does not support 8bpp mode.
		// Use FadeOut()
	}
	else
	{
		// Create a tmp surface the size of the front buffer and
		// copy the front buffer to it.
		w = GetBack()->GetWidth();
		h = GetBack()->GetHeight();

		// Need to get the RECT defining the CDX app window
		GetClientRect((HWND)m_hWnd, &sRect);
		pt.x = pt.y = 0;
		ClientToScreen((HWND)m_hWnd, &pt);
		OffsetRect(&sRect, pt.x, pt.y);

		// Create the temp buffer
		tmp = new CDXSurface(this, w, h, CDXMEM_SYSTEMONLY);

		GetFront()->CDXBLT_Blk(tmp, 0, 0, sRect);

		ALPHA = 256;
		sRect.left   = 0;
		sRect.top    = 0;
		sRect.right  = w;
		sRect.bottom = h;

		// Continue fading while decrementing the ALPHA value.
		while(ALPHA > 0)
		{
			fadeBegin = timeGetTime();

			tmp->CDXBLT_Fade(GetBack(),0,0,sRect,(WORD)ALPHA);
			Flip(TRUE,FALSE);

			fadeInc = (float)(duration + 1) / (float)(timeGetTime() - fadeBegin);

			// Do some boundry checking
			if (fadeInc > 256)
				fadeInc = 256;
			else if (fadeInc < 1)
				fadeInc = 1;

			ALPHA -= (LONG)(512 / fadeInc);
		}

		// Cleanup the surface
		delete tmp;
		tmp = NULL;

		// Just to be sure we have a pure BLACK screen
		GetBack()->Fill(0);
		Flip(TRUE,FALSE);
	}*/
}

/************************************************************************************/
// ALPHA OPERACE
/************************************************************************************/

//////////////////////////////////////////////////////////////////////////////////
// Inicializuje alphablitting - v CreateFullScreen
//////////////////////////////////////////////////////////////////////////////////
HRESULT HDDraw::InitAlpha(DDPIXELFORMAT& r_pixelFormat)
{
	m_rMask = r_pixelFormat.dwRBitMask;
   m_gMask = r_pixelFormat.dwGBitMask;
   m_bMask = r_pixelFormat.dwBBitMask;

   // compute darken50 and-mask
   unsigned long rMask = m_rMask;
   unsigned long gMask = m_gMask;
   unsigned long bMask = m_bMask;
   unsigned long x;

   x = 1; while ((rMask & x) == 0) x <<= 1; rMask ^= x;
   x = 1; while ((gMask & x) == 0) x <<= 1; gMask ^= x;
   x = 1; while ((bMask & x) == 0) x <<= 1; bMask ^= x;

   m_FadeMask = rMask | gMask | bMask;

   // success
   return DD_OK;
}

//////////////////////////////////////////////////////////////////////////////////
// ztmavi jeden blok
//////////////////////////////////////////////////////////////////////////////////
HRESULT HDDraw::FadeBlockC(
        DDSURFACEDESC& r_destSurface,
        int destX,
        int destY,
        int width,
        int height)
{
	if (width == 0 || height == 0) return DD_OK;

   // compute pointers and modulo
   unsigned short * p_bits = (unsigned short *)&(((unsigned char *)r_destSurface.lpSurface)[(destX << 1) + (destY * r_destSurface.lPitch)]);
   unsigned long modulo = r_destSurface.lPitch - (width << 1);

   // perform the alpha-blit
   unsigned long x;
   unsigned long mask = m_FadeMask;

   do {
		x = width;
      do {
         *p_bits++ = (unsigned short)((*p_bits & mask) >> 1);
      } while (--x != 0);
	
      p_bits = (unsigned short * )&(((byte *)p_bits)[modulo]);
	
   } while (--height != 0);

   return DD_OK;
}

//////////////////////////////////////////////////////////////////////////////////
// kopiruje polopruhledne - zalezi na promene alpha
//////////////////////////////////////////////////////////////////////////////////
HRESULT HDDraw::AlphaBltC(
        DDSURFACEDESC& r_destSurface,
        int destX, 
        int destY,
        DDSURFACEDESC& r_srcSurface,
        RECT * p_srcRect,
        byte alpha)
{
	if (alpha == 0) return S_OK;

   // get blitting dimensions
   int srcX = 0;
   int srcY = 0;
   unsigned long  width  = r_srcSurface.dwWidth;
   unsigned long height = r_srcSurface.dwHeight;

   if (p_srcRect != NULL) {
      width  = p_srcRect->right;
      height = p_srcRect->bottom;
      srcX   = p_srcRect->left;
      srcY   = p_srcRect->top;
   }
        if (width == 0 || height == 0) return S_OK;

        // compute pointers and modulo
        unsigned short * p_destBits = (unsigned short *)&(((unsigned char *)r_destSurface.lpSurface)[(destX << 1) + (destY * r_destSurface.lPitch)]);
        unsigned short * p_srcBits  = (unsigned short *)&(((unsigned char *)r_srcSurface.lpSurface)[(srcX << 1) + (srcY * r_srcSurface.lPitch)]);
        unsigned long destModulo = r_destSurface.lPitch - (width << 1);
        unsigned long srcModulo  = r_srcSurface.lPitch - (width << 1);

        // perform the alpha-blit
        unsigned long x;
        unsigned long srcPixel, destPixel;

        unsigned long rMask = m_rMask;
        unsigned long gMask = m_gMask;
        unsigned long bMask = m_bMask;
        unsigned long r, g, b;

        do
        {
                x = width;
                do
                {
                        // read the pixel from the source surface
                        srcPixel = *p_srcBits++;

                        // read the pixel from the destination surface
                        destPixel = *p_destBits;

                        // blend color channels
                        r = destPixel & rMask;
                        g = destPixel & gMask;
                        b = destPixel & bMask;

                        destPixel  = (((((srcPixel & rMask) - r) * alpha) >> 8) & rMask) + r;
                        destPixel |= (((((srcPixel & gMask) - g) * alpha) >> 8) & gMask) + g;
                        destPixel |= (((((srcPixel & bMask) - b) * alpha) >> 8) & bMask) + b;

                        // write pixel
                        *p_destBits++ = (unsigned short)destPixel;

                } while (--x != 0);

                p_srcBits  = (unsigned short * )&(((unsigned char *)p_srcBits)[srcModulo]);
                p_destBits = (unsigned short * )&(((unsigned char *)p_destBits)[destModulo]);

        } while (--height != 0);

        // success
        return DD_OK;
}

//////////////////////////////////////////////////////////////////////////////////
// kopiruje polopruhledne - podle masky - musi bit 256 bitmapa ze stupnu sede
//////////////////////////////////////////////////////////////////////////////////
HRESULT HDDraw::AlphaBltExC(
        DDSURFACEDESC& r_destSurface,
        int destX, 
        int destY,
        DDSURFACEDESC& r_srcSurface,
        RECT * p_srcRect,
        DDSURFACEDESC& r_alphaSurface)
{
        // get blitting dimensions
   int srcX = 0;
   int srcY = 0;
   unsigned long width  = r_srcSurface.dwWidth;
   unsigned long height = r_srcSurface.dwHeight;

   if (p_srcRect != NULL)  {
      width  = p_srcRect->right;
      height = p_srcRect->bottom;
      srcX   = p_srcRect->left;
      srcY   = p_srcRect->top;
   }

   if (width == 0 || height == 0) return DD_OK;

   // compute pointers and modulo
   unsigned short * p_destBits  = (unsigned short *)&(((unsigned char *)r_destSurface.lpSurface)[(destX << 1) + (destY * r_destSurface.lPitch)]);
   unsigned short * p_srcBits   = (unsigned short *)&(((unsigned char *)r_srcSurface.lpSurface)[(srcX << 1) + (srcY * r_srcSurface.lPitch)]);
   byte   * p_alphaBits = (unsigned char *) &(((unsigned char*)r_alphaSurface.lpSurface)[srcX + srcY * r_alphaSurface.lPitch]);
   unsigned long destModulo  = r_destSurface.lPitch - (width << 1);
   unsigned long srcModulo   = r_srcSurface.lPitch - (width << 1);
   unsigned long alphaModulo = r_alphaSurface.lPitch - width;

   // perform the alpha-blit
   unsigned long x;
   unsigned long srcPixel, destPixel;

   unsigned long rMask = m_rMask;
   unsigned long gMask = m_gMask;
   unsigned long bMask = m_bMask;
   unsigned long r, g, b;
   unsigned long alpha;

   do {
      x = width;
      do {
         // read the pixel from the source surface
         srcPixel = *p_srcBits++;
         alpha = *p_alphaBits++;

         // read the pixel from the destination surface
         destPixel = *p_destBits;

         // blend color channels
         r = destPixel & rMask;
         g = destPixel & gMask;
         b = destPixel & bMask;

         destPixel  = (((((srcPixel & rMask) - r) * alpha) >> 8) & rMask) + r;
         destPixel |= (((((srcPixel & gMask) - g) * alpha) >> 8) & gMask) + g;
         destPixel |= (((((srcPixel & bMask) - b) * alpha) >> 8) & bMask) + b;

         // write pixel
         *p_destBits++ = (unsigned short)destPixel;

       } while (--x != 0);

       p_srcBits   = (unsigned short * )&(((unsigned char *)p_srcBits)[srcModulo]);
       p_destBits  = (unsigned short * )&(((unsigned char *)p_destBits)[destModulo]);
       p_alphaBits = p_alphaBits + alphaModulo;

   } while (--height != 0);

   // success
   return DD_OK;
}


HRESULT HDDraw::FadeBlock  (LPDIRECTDRAWSURFACE lpdest, int destX, int destY, int width, int height)
{
	HRESULT ret;
	DDSURFACEDESC  src;
	DDSURFACEDESC  dest;

	DDInitStruct(&src);
	DDInitStruct(&dest);

	lpdest->Lock(NULL, &dest, DDLOCK_WAIT, NULL);

	ret = FadeBlockC(dest, destX, destY, width, height);

	lpdest->Unlock(NULL);

   return ret;
}

HRESULT HDDraw::AlphaBlt   (LPDIRECTDRAWSURFACE lpdest, int destX, int destY, LPDIRECTDRAWSURFACE lpsrc, RECT * p_rect, byte alpha)
{
	HRESULT ret;
	DDSURFACEDESC  src;
	DDSURFACEDESC  dest;

	DDInitStruct(&src);
	DDInitStruct(&dest);

	lpsrc->Lock(NULL, &src, DDLOCK_WAIT, NULL);
	lpdest->Lock(NULL, &dest, DDLOCK_WAIT, NULL);

	ret = AlphaBltC(dest, destX, destY, src, p_rect, alpha);

	lpsrc->Unlock(NULL);
	lpdest->Unlock(NULL);

   return ret;
}

HRESULT HDDraw::AlphaBltEx (LPDIRECTDRAWSURFACE lpdest, int destX, int destY, LPDIRECTDRAWSURFACE lpsrc, RECT * p_rect, HBITMAP halpha)
{
	HRESULT ret;
	DDSURFACEDESC  src;
	DDSURFACEDESC  dest;
	DDSURFACEDESC  alpha;

	DDInitStruct(&src);
	DDInitStruct(&dest);
	DDInitStruct(&alpha);

	lpsrc->Lock(NULL, &src, DDLOCK_WAIT, NULL);
	lpdest->Lock(NULL, &dest, DDLOCK_WAIT, NULL);

	BITMAP bm;
	GetObject(halpha, sizeof(bm), &bm);
   alpha.dwWidth   = bm.bmWidth;
   alpha.dwHeight  = bm.bmHeight;
   alpha.lPitch    = bm.bmWidthBytes;
   alpha.lpSurface = bm.bmBits;

	ret = AlphaBltExC(dest, destX, destY, src, p_rect, alpha);

	lpsrc->Unlock(NULL);
	lpdest->Unlock(NULL);

   return ret;
}

//////////////////////////////////////////////////////////////////////////////////
// umoznuje pouzivat tridu i jako LPDIRECTDRAW
//////////////////////////////////////////////////////////////////////////////////
HDDraw::operator LPDIRECTDRAW() {
	return m_lpDD;
}

//////////////////////////////////////////////////////////////////////////////////
// Toto volejte pred zobrazovanim dialogu ve Full Screenu
//////////////////////////////////////////////////////////////////////////////////
HRESULT HDDraw::FlipToGDI()
{
	if(!m_GDIVisible) return Flip();
	return DD_OK;
}
//////////////////////////////////////////////////////////////////////////////////
// Pokud potrebujete psat, nebo kreslit na back surface
//////////////////////////////////////////////////////////////////////////////////
HDC HDDraw::GetBackDC()
{
	m_lpDDSBack->GetDC(&m_dcBack);
	return m_dcBack;
}
//////////////////////////////////////////////////////////////////////////////////
// kdyz dokreslite, tak zavoleste tohle - jinak nepude surface pouzit !!!
//////////////////////////////////////////////////////////////////////////////////
void HDDraw::ReleaseBackDC()
{
	m_lpDDSBack->ReleaseDC(m_dcBack);
//	m_dcBack = NULL;
}
//////////////////////////////////////////////////////////////////////////////////
// Pokud potrebujete psat, nebo kreslit na front surface
//////////////////////////////////////////////////////////////////////////////////
HDC HDDraw::GetGDIDC()
{
	m_lpDDSGDI->GetDC(&m_dcGDI);
	return m_dcGDI;
}
//////////////////////////////////////////////////////////////////////////////////
// kdyz dokreslite, tak zavoleste tohle - jinak nepude surface pouzit !!!
//////////////////////////////////////////////////////////////////////////////////
void HDDraw::ReleaseGDIDC()
{
	m_lpDDSGDI->ReleaseDC(m_dcGDI);
//	m_dcBack = NULL;
}
